#! /usr/bin/env python
# -*- coding: UTF-8 -*-
import tkinter as Tk
from tkinter import scrolledtext as S
from tkinter import filedialog as D
import os
class Frame(Tk.Frame):
    def __init__(self, master=None):
        Tk.Frame.__init__(self, master)
        self.master.title(os.name=='posix' and 'untitled' or '無題')
        self.fileName = None  # ファイル名
        self.retWin   = None  # 検索/置換用ダイアログボックス
        #---- Menu -----------------------------------------------------------------------
        menuBar = Tk.Menu(self, tearoff=0)
        # File
        menuFile = Tk.Menu(menuBar, tearoff=0)
        menuBar.add_cascade(label="ファイル(F)", menu=menuFile,  underline=5)
        menuFile.add_command(label="新規作成(N)", command=self.newProc, underline=5, accelerator = 'Ctrl-N')
        menuFile.add_command(label="開く(O)", command=self.openProc, underline=3, accelerator = 'Ctrl-O')
        menuFile.add_command(label="保存(S)", command=self.saveProc, underline=3, accelerator = 'Ctrl-S')
        menuFile.add_command(label="名前をつけて保存(A)", command=self.saveasProc, underline=9)
        menuFile.add_separator()
        menuFile.add_command(label="終了(Q)", command=self.exit, underline=3 , accelerator = 'Ctrl-Q')
        # Edit
        menuEdit = Tk.Menu(menuBar, tearoff=0)
        menuBar.add_cascade(label='編集(E)', menu=menuEdit, underline=3)
        menuEdit.add_command(label='全てを選択(A)', command=self.selectAll, underline=6,
                             accelerator = 'Ctrl-A')
        menuEdit.add_command(label='切り取り(X)', command=self.cut, underline=5,
                             accelerator = os.name=='posix' and 'Ctrl-W' or 'Ctrl-X')
        menuEdit.add_command(label='コピー(C)', command=self.copy, underline=4,
                             accelerator = os.name=='posix' and 'Alt-W' or 'Ctrl-C')
        menuEdit.add_command(label='貼り付け(V)', command=self.paste, underline=5,
                             accelerator = os.name=='posix' and 'Ctrl-Y' or 'Ctrl-V')
        menuEdit.add_separator()
        menuEdit.add_command(label='行選択', command=self.selectLine, accelerator = 'Ctrl-L')
        menuEdit.add_command(label='行削除', command=self.deleteLine, accelerator = 'Ctrl-G')
        menuEdit.add_separator()
        menuEdit.add_command(label='検索(F)', command=self.find, underline=3, accelerator = 'Ctrl-F')
        menuEdit.add_command(label='置換(R)', command=self.replace, underline=3, accelerator = 'Ctrl-R')
        # Tool
        menuTool = Tk.Menu(menuBar, tearoff=0)
        menuBar.add_cascade(label='ツール(T)', menu=menuTool, underline=3)
        menuTool.add_command(label='実行(E)', command=self.execute, underline=3, accelerator = 'Ctrl-E')
        menuTool.add_command(label='Eval(I)', command=self.eval, underline=6, accelerator = 'Ctrl-I')
        # short-cuts
        self.master.bind('<Control-KeyPress-o>', self.openProc)
        self.master.bind('<Control-KeyPress-s>', self.saveProc)
        self.master.bind('<Control-KeyPress-q>', self.exit)
        self.master.bind('<Control-KeyPress-a>', self.selectAll)
        self.master.bind('<Control-KeyPress-e>', self.execute)
        self.master.bind('<Control-KeyPress-l>', self.selectLine)
        self.master.bind('<Control-KeyPress-g>', self.deleteLine)
        self.master.bind('<Control-KeyPress-i>', self.eval)
        self.master.bind('<Control-KeyPress-f>', self.find)
        self.master.bind('<Control-KeyPress-r>', self.replace)
        self.master.bind('<Shift-KeyPress-Delete>', self.deleteLine)
        self.master.bind('<Double-Button-1>', self.deleteLine)
        if os.name == 'posix': self.master.bind('<Alt-KeyPress-w>', self.copy)
        try:
            self.master.config(menu=menuBar)
        except AttributeError:
            self.master.Tk.call(master, "config", "-menu", menuBar)
        # テキストボックス(スクロールバー付き)
        self.scV = Tk.Scrollbar(self,orient=Tk.VERTICAL) 
        self.scH = Tk.Scrollbar(self,orient=Tk.HORIZONTAL ) 
        self.txt = Tk.Text(self, wrap=Tk.NONE,
                           yscrollcommand=self.scV.set, xscrollcommand=self.scH.set,
                           borderwidth=0, highlightthickness=0,
                           font=('sans-serif', '10'))
        self.scV.config(command=self.txt.yview);self.scV.pack(side="right", fill="y")
        self.scH.config(command=self.txt.xview);self.scH.pack(side="bottom", fill="x")
        self.txt.pack(fill="both", expand=True)
        self.txt.focus_set()
    #--------------------------------------------------------
    #■ 検索/置換用
    #　 共通コントロール・グループ 1
    def findReplace01(self,Title):
        self.RefID=0
        self.S1 = "" #検索結果位置なしとする
        self.txt.tag_add("here", "1.0","1.0")
        self.retWin=Tk.Toplevel()
        self.retWin.title(Title)
        self.retWin.geometry("300x60")
        #検索用文字列
        self.text1=Tk.Text(self.retWin, width=40, height=1); self.text1.pack()
        self.text1.focus_set()
    #   共通コントロール・グループ 2
    def findReplace02(self):
        self.findMsg=Tk.StringVar() # 検索失敗表示用ラベル
        self.label1=Tk.Label(self.retWin,textvariable=self.findMsg)
        self.label1.pack(anchor='sw',side='left')
        # 検索用ボタン
        Tk.Button(self.retWin,text="取消", command=self.findBoxEnd).pack(anchor='ne',side='right')
        Tk.Button(self.retWin,text="検索", command=self.findExec  ).pack(anchor='ne',side='right')
        self.retWin.protocol("WM_DELETE_WINDOW",self.findBoxEnd)
    #■検索用ダイアログボックス
    def findBox(self):
        if self.retWin is None or not retWin.winfo_exists():
            self.findReplace01("検索");self.findReplace02()
    #■置換用ダイアログボックス
    def replaceBox(self):
        if self.retWin is None or not retWin.winfo_exists():
            self.findReplace01("置換")                            # 置換/検索共通
            self.text2=Tk.Text(self.retWin, width=40, height=1)   # 置換用テキストボックス
            self.text2.pack()
            self.findReplace02()                                 # 置換/検索共通
            Tk.Button(self.retWin,text="置換"  , command=self.replaceExec    ).pack(anchor='ne',side='right')
    #■検索/置換キャンセル
    def findBoxEnd(self, event=None):
        self.txt.tag_config("here", background='white')
        self.txt.tag_delete("here");  self.S1=""
        self.retWin.destroy(); self.retWin=None
    #■置換実行
    def replaceExec(self, event=None) :
        if self.S1=="" : return                  # 検索文字列なしのとき無視
        self.txt.delete(self.S1,self.S2)         # 検索文字列を削除
        rs1=self.text1.get('1.0', Tk.END+'-1c')  #    検索文字列
        rs2=self.text2.get('1.0', Tk.END+'-1c')  #    置換文字列
        self.txt.insert(self.S1,rs2)             # 置換文字列を挿入
        self.RefID+=len(rs2)-len(rs1)            # 検索開始位置をずらす
        self.findExec(None)                      # 次の文字列検索
    #■検索実行
    def findExec(self, event=None):
        self.findMsg.set("")
        rs=self.text1.get('1.0', Tk.END+'-1c')#検索文字列の取り出し
        if rs=="":return
        self.txt.tag_delete("here");  self.S1=""
        dt=self.txt.get('1.0', Tk.END)
        i=0; Ns=len(rs); N=len(dt);j=1
        for k in range(N):
            if dt[k]=='\n': j+=1; i=0 # 行末のとき行カウント
            elif k<self.RefID: i += 1 # 検索開始位置前なら桁位置カウントのみ
            elif dt[k]==rs[0]:        # 検索文字列の先頭文字に等しければ比較開始
                flag=True
                for kk in range(1,Ns):
                    ii=k+kk
                    if ii>=N: flag=False; break    # 被検索文字列の最後のとき失敗
                    if dt[ii]!=rs[kk]: flag=False  # 検索文字列と等しくない文字のとき失敗
                if flag: # 最後まで等しければ成功とする。(背景色を変える)
                    self.RefID=k+Ns+1; ied=i+Ns
                    self.S1="%d.%d" % (j,i); self.S2="%d.%d" % (j,ied)
                    self.txt.mark_set(Tk.SEL,self.S2) # 検索位置を見えるようにする
                    self.txt.see(Tk.SEL)
                    self.txt.tag_add("here", self.S1,self.S2)
                    self.txt.tag_config("here", background='yellow')
                    return
                i+=1
            else: i+=1
        self.findMsg.set("文字列が見付かりませんでした")
        self.S1="";  self.RefID=0
    #----------(以下、各コマンド処理)----------------------------------------
    def newProc(self, event=None):
        self.fileName=None
        self.master.title(os.name=='posix' and 'untitled' or '無題')
        self.txt.delete('1.0', Tk.END)

    def openProc(self, event=None):
        fname = D.askopenfilename(filetypes =[('text files', '*.txt'), ('all files', '*.*')])
        if fname:
            self.txt.delete('1.0', Tk.END)
            fp=open(fname,'rt')
            self.txt.insert(Tk.END, fp.read())
            fp.close()
            self.fileName = fname
            self.master.title(fname)
         
    def save(self, fp):
        fp.write(self.txt.get('1.0', Tk.END))
        fp.close()
                  
    def saveProc(self, event=None):
        if self.fileName:
            self.save(open(self.fileName, 'wt'))
        else:
            self.saveasProc()
     
    def saveasProc(self):
        fname = D.asksaveasfilename(filetypes =[('text files', '*.txt')])
        if fname:
            self.save(open(fname, 'wt'))
            self.fileName=fname
            self.master.title(fname)

    def selectAll(self, event=None):
        self.txt.tag_add(Tk.SEL, '1.0', Tk.END+'-1c')
        self.txt.mark_set(Tk.INSERT, '1.0')
        self.txt.see(Tk.INSERT)

    def cut(self, event=None):
        if self.txt.tag_ranges(Tk.SEL):
            self.copy()
            self.txt.delete(Tk.SEL_FIRST, Tk.SEL_LAST)

    def copy(self, event=None):
        if self.txt.tag_ranges(Tk.SEL):
            text = self.txt.get(Tk.SEL_FIRST, Tk.SEL_LAST)
            self.clipboard_clear()
            self.clipboard_append(text)

    def paste(self, event=None):
        text = self.selection_get(selection='CLIPBOARD')
        if text:
            self.txt.insert(Tk.INSERT, text)
            self.txt.tag_remove(Tk.SEL, '1.0', Tk.END)
            self.txt.see(Tk.INSERT)

    def selectLine(self, event=None):
        self.txt.tag_add(Tk.SEL, Tk.INSERT + " linestart", Tk.INSERT + " lineend")
        self.txt.mark_set(Tk.INSERT, Tk.INSERT + " linestart")
        self.txt.see(Tk.INSERT)

    def deleteLine(self, event=None):
        self.txt.delete(Tk.INSERT + " linestart", Tk.INSERT + " lineend")

    def execute(self, event=None): exec(self.txt.get('1.0', Tk.END))
        
    def eval(self, event=None):
        S=self.txt.get('1.0', Tk.END)
        Es=""; k=1; Dict={}
        for i in range(len(S)):
            if S[i]=='\n':
                if Es!='' and Es!='\n':#行の終りでEval
                    print(k," : ", Es)
                    try:
                        N=len(Es)-1; ID=0; j=0; inStr=0 #代入文の判定
                        while j<N-1: # '='が1個のとき代入文。'=='は等値演算子
                            if Es[j] =='\\' : j += 1     #次の1文字無視
                            elif inStr==1 and Es[j] =='\"': inStr=0 # 文字列終了
                            elif inStr==2 and Es[j] =='\'': inStr=0 # 文字列終了
                            elif inStr==0 and Es[j] =='\"': inStr=1 # 二重引用符内開始
                            elif inStr==0 and Es[j] =='\'': inStr=2 # 引用符内開始
                            elif inStr==0 and Es[j] =='=':
                                if Es[j+1]!='=':ID=j; break    # 1個の'='で代入文
                                j+=1 # ==のとき
                            j+=1
                        if ID!=0: # (先頭に '=' がくることはないのでエラーでもよい)
                            vstr=""; dstr=""                          #代入文の処理
                            for j in range(ID):                       #変数名セット
                                if Es[j]!=' ' : vstr+=Es[j]
                            for j in range(ID+1, len(Es)):dstr+=Es[j] #右辺式セット
                            Dict[vstr]=eval(dstr,Dict)
                        else: eval(Es, Dict)
                    except: print(k, ":","エラーが起きました。<"+Es+">")
                    Es=''; k+=1
            else : Es+=S[i]
             
    def find(self, event=None): self.findBox()

    def replace(self, event=None):self.replaceBox()
        
    def exit(self, event=None): self.master.destroy()
#----------------------------------------------------------
#■メイン処理 
if __name__ == '__main__':
    frame = Frame()
    frame.pack()
    frame.mainloop()
